@@ -19,22 +19,26 @@ from autoapi.mappers.python import (
 )
 
 
+def rebuild(confoverrides=None, **kwargs):
+    app = Sphinx(
+        srcdir=".",
+        confdir=".",
+        outdir="_build/text",
+        doctreedir="_build/.doctrees",
+        buildername="text",
+        confoverrides=confoverrides,
+        **kwargs
+    )
+    app.build()
+
+
 @pytest.fixture(scope="class")
 def builder():
     cwd = os.getcwd()
 
     def build(test_dir, confoverrides=None, **kwargs):
         os.chdir("tests/python/{0}".format(test_dir))
-        app = Sphinx(
-            srcdir=".",
-            confdir=".",
-            outdir="_build/text",
-            doctreedir="_build/.doctrees",
-            buildername="text",
-            confoverrides=confoverrides,
-            **kwargs
-        )
-        app.build(force_all=True)
+        rebuild(confoverrides=confoverrides, **kwargs)
 
     yield build
 
@@ -60,8 +64,7 @@ class TestSimpleModule(object):
         with io.open(example_path, encoding="utf8") as example_handle:
             example_file = example_handle.read()
 
-        if sphinx.version_info >= (2,):
-            assert "@example.decorator_okay" in example_file
+        assert "@example.decorator_okay" in example_file
 
     def check_integration(self, example_path):
         with io.open(example_path, encoding="utf8") as example_handle:
@@ -114,9 +117,6 @@ class TestSimpleModule(object):
         assert "Bases:" in example_file
 
 
-@pytest.mark.skipif(
-    sys.version_info < (3,), reason="Ellipsis is invalid method contents in Python 2"
-)
 class TestSimpleStubModule(object):
     @pytest.fixture(autouse=True, scope="class")
     def built(self, builder):
@@ -158,9 +158,6 @@ class TestSimpleStubModuleNotPreferred(object):
         assert "Foo" in example_file
 
 
-@pytest.mark.skipif(
-    sys.version_info < (3, 6), reason="Annotations are invalid in Python <3.5"
-)
 class TestPy3Module(object):
     @pytest.fixture(autouse=True, scope="class")
     def built(self, builder):
@@ -192,8 +189,9 @@ class TestPy3Module(object):
 
         assert "global_a :A" in example_file
 
-        if sphinx.version_info >= (2, 1):
-            assert "my_method(self) -> str" in example_file
+        assert "my_method(self) -> str" in example_file
+
+        assert "class example.SomeMetaclass" in example_file
 
     def test_overload(self):
         example_path = "_build/text/autoapi/example/index.txt"
@@ -223,17 +221,10 @@ class TestPy3Module(object):
         with io.open(example_path, encoding="utf8") as example_handle:
             example_file = example_handle.read()
 
-        if sphinx.version_info >= (2, 1):
-            assert "async async_method" in example_file
-            assert "async example.async_function" in example_file
-        else:
-            assert "async_method" in example_file
-            assert "async_function" in example_file
+        assert "async async_method" in example_file
+        assert "async example.async_function" in example_file
 
 
-@pytest.mark.skipif(
-    sys.version_info < (3, 6), reason="Annotations are invalid in Python <3.5"
-)
 def test_py3_hiding_undoc_overloaded_members(builder):
     confoverrides = {"autoapi_options": ["members", "special-members"]}
     builder("py3example", confoverrides=confoverrides)
@@ -248,9 +239,6 @@ def test_py3_hiding_undoc_overloaded_members(builder):
     assert "undoc_overloaded_method" not in example_file
 
 
-@pytest.mark.skipif(
-    sys.version_info < (3,), reason="Annotations are not supported in astroid<2"
-)
 class TestAnnotationCommentsModule(object):
     @pytest.fixture(autouse=True, scope="class")
     def built(self, builder):
@@ -702,9 +690,51 @@ class TestComplexPackageParallel(object):
         builder("pypackagecomplex", parallel=2)
 
 
-@pytest.mark.skipif(
-    sys.version_info < (3, 3), reason="Implicit namespace not supported in python < 3.3"
-)
+def test_caching(builder):
+    mtimes = (0, 0)
+
+    def record_mtime():
+        nonlocal mtimes
+        mtime = 0
+        for root, _, files in os.walk("_build/text/autoapi"):
+            for name in files:
+                this_mtime = os.path.getmtime(os.path.join(root, name))
+                mtime = max(mtime, this_mtime)
+
+        mtimes = (*mtimes[1:], mtime)
+
+    builder("pypackagecomplex", confoverrides={"autoapi_keep_files": True})
+    record_mtime()
+
+    rebuild(confoverrides={"autoapi_keep_files": True})
+    record_mtime()
+
+    assert mtimes[1] == mtimes[0]
+
+    # Check that adding a file rebuilds the docs
+    extra_file = "complex/new.py"
+    with open(extra_file, "w") as out_f:
+        out_f.write("\n")
+
+    try:
+        rebuild(confoverrides={"autoapi_keep_files": True})
+    finally:
+        os.remove(extra_file)
+
+    record_mtime()
+    assert mtimes[1] != mtimes[0]
+
+    # Removing a file also rebuilds the docs
+    rebuild(confoverrides={"autoapi_keep_files": True})
+    record_mtime()
+    assert mtimes[1] != mtimes[0]
+
+    # Changing not keeping files always builds
+    rebuild()
+    record_mtime()
+    assert mtimes[1] != mtimes[0]
+
+
 class TestImplicitNamespacePackage(object):
     @pytest.fixture(autouse=True, scope="class")
     def built(self, builder):
@@ -723,3 +753,24 @@ class TestImplicitNamespacePackage(object):
             example_file = example_handle.read()
 
         assert "namespace.example.second_sub_method" in example_file
+
+
+def test_custom_jinja_filters(builder):
+    confoverrides = {
+        "autoapi_prepare_jinja_env": (
+            lambda jinja_env: jinja_env.filters.update(
+                {
+                    "prepare_docstring": (
+                        lambda docstring: "This is using custom filters."
+                    )
+                }
+            )
+        ),
+    }
+    builder("pyexample", confoverrides=confoverrides)
+
+    example_path = "_build/text/autoapi/example/index.txt"
+    with io.open(example_path, encoding="utf8") as example_handle:
+        example_file = example_handle.read()
+
+    assert "This is using custom filters." in example_file
